@@ -12,8 +12,9 @@
 # language governing permissions and limitations under the License.
 import copy
 import logging
-from collections import defaultdict, deque, namedtuple
+from collections import deque, namedtuple
 from botocore.compat import accepts_kwargs, six
+from botocore.utils import EVENT_ALIASES
 
 logger = logging.getLogger(__name__)
 
@@ -23,6 +24,7 @@ _FIRST = 0
 _MIDDLE = 1
 _LAST = 2
 
+
 class NodeList(_NodeList):
 
     def __copy__(self):
@@ -170,7 +172,7 @@ class HierarchicalEmitter(BaseEventHooks):
         self._handlers = _PrefixTrie()
         # This is used to ensure that unique_id's are only
         # registered once.
-        self._unique_id_cache = {}
+        self._unique_id_handlers = {}
 
     def _emit(self, event_name, kwargs, stop_on_response=False):
         """
@@ -259,40 +261,40 @@ class HierarchicalEmitter(BaseEventHooks):
         self._register_section(event_name, handler, unique_id,
                                unique_id_uses_count, section=_LAST)
 
-
     def _register_section(self, event_name, handler, unique_id,
                           unique_id_uses_count, section):
         if unique_id is not None:
-            if unique_id in self._unique_id_cache:
+            if unique_id in self._unique_id_handlers:
                 # We've already registered a handler using this unique_id
                 # so we don't need to register it again.
-                count = self._unique_id_cache[unique_id].get('count', None)
+                count = self._unique_id_handlers[unique_id].get('count', None)
                 if unique_id_uses_count:
                     if not count:
-                        raise ValueError("Initial registration of"
-                            " unique id %s was specified to use a counter."
-                            " Subsequent register calls to unique id must"
-                            " specify use of a counter as well." % unique_id)
+                        raise ValueError(
+                            "Initial registration of  unique id %s was "
+                            "specified to use a counter. Subsequent register "
+                            "calls to unique id must specify use of a counter "
+                            "as well." % unique_id)
                     else:
-                        self._unique_id_cache[unique_id]['count'] += 1
+                        self._unique_id_handlers[unique_id]['count'] += 1
                 else:
                     if count:
-                        raise ValueError("Initial registration of"
-                            " unique id %s was specified to not use a counter."
-                            " Subsequent register calls to unique id must"
-                            " specify not to use a counter as well." %
-                            unique_id)
+                        raise ValueError(
+                            "Initial registration of unique id %s was "
+                            "specified to not use a counter. Subsequent "
+                            "register calls to unique id must specify not to "
+                            "use a counter as well." % unique_id)
                 return
             else:
                 # Note that the trie knows nothing about the unique
                 # id.  We track uniqueness in this class via the
-                # _unique_id_cache.
+                # _unique_id_handlers.
                 self._handlers.append_item(event_name, handler,
                                            section=section)
-                unique_id_cache_item = {'handler': handler}
+                unique_id_handler_item = {'handler': handler}
                 if unique_id_uses_count:
-                    unique_id_cache_item['count'] = 1
-                self._unique_id_cache[unique_id] = unique_id_cache_item
+                    unique_id_handler_item['count'] = 1
+                self._unique_id_handlers[unique_id] = unique_id_handler_item
         else:
             self._handlers.append_item(event_name, handler, section=section)
         # Super simple caching strategy for now, if we change the registrations
@@ -303,30 +305,30 @@ class HierarchicalEmitter(BaseEventHooks):
                    unique_id_uses_count=False):
         if unique_id is not None:
             try:
-                count = self._unique_id_cache[unique_id].get('count', None)
+                count = self._unique_id_handlers[unique_id].get('count', None)
             except KeyError:
                 # There's no handler matching that unique_id so we have
                 # nothing to unregister.
                 return
             if unique_id_uses_count:
-                if count == None:
-                    raise ValueError("Initial registration of"
-                        " unique id %s was specified to use a counter."
-                        " Subsequent unregister calls to unique id must"
-                        " specify use of a counter as well." % unique_id)
+                if count is None:
+                    raise ValueError(
+                        "Initial registration of unique id %s was specified to "
+                        "use a counter. Subsequent unregister calls to unique "
+                        "id must specify use of a counter as well." % unique_id)
                 elif count == 1:
-                    handler = self._unique_id_cache.pop(unique_id)['handler']
+                    handler = self._unique_id_handlers.pop(unique_id)['handler']
                 else:
-                    self._unique_id_cache[unique_id]['count'] -= 1
+                    self._unique_id_handlers[unique_id]['count'] -= 1
                     return
             else:
                 if count:
-                    raise ValueError("Initial registration of"
-                        " unique id %s was specified to not use a counter."
-                        " Subsequent unregister calls to unique id must"
-                        " specify not to use a counter as well." % 
-                        unique_id)
-                handler = self._unique_id_cache.pop(unique_id)['handler']
+                    raise ValueError(
+                        "Initial registration of unique id %s was specified "
+                        "to not use a counter. Subsequent unregister calls "
+                        "to unique id must specify not to use a counter as "
+                        "well." % unique_id)
+                handler = self._unique_id_handlers.pop(unique_id)['handler']
         try:
             self._handlers.remove_item(event_name, handler)
             self._lookup_cache = {}
@@ -337,10 +339,109 @@ class HierarchicalEmitter(BaseEventHooks):
         new_instance = self.__class__()
         new_state = self.__dict__.copy()
         new_state['_handlers'] = copy.copy(self._handlers)
+        new_state['_unique_id_handlers'] = copy.copy(self._unique_id_handlers)
         new_instance.__dict__ = new_state
         return new_instance
 
 
+class EventAliaser(BaseEventHooks):
+    def __init__(self, event_emitter, event_aliases=None):
+        self._event_aliases = event_aliases
+        if event_aliases is None:
+            self._event_aliases = EVENT_ALIASES
+        self._alias_name_cache = {}
+        self._emitter = event_emitter
+
+    def emit(self, event_name, **kwargs):
+        aliased_event_name = self._alias_event_name(event_name)
+        return self._emitter.emit(aliased_event_name, **kwargs)
+
+    def emit_until_response(self, event_name, **kwargs):
+        aliased_event_name = self._alias_event_name(event_name)
+        return self._emitter.emit_until_response(aliased_event_name, **kwargs)
+
+    def register(self, event_name, handler, unique_id=None,
+                 unique_id_uses_count=False):
+        aliased_event_name = self._alias_event_name(event_name)
+        return self._emitter.register(
+            aliased_event_name, handler, unique_id, unique_id_uses_count
+        )
+
+    def register_first(self, event_name, handler, unique_id=None,
+                       unique_id_uses_count=False):
+        aliased_event_name = self._alias_event_name(event_name)
+        return self._emitter.register_first(
+            aliased_event_name, handler, unique_id, unique_id_uses_count
+        )
+
+    def register_last(self, event_name, handler, unique_id=None,
+                      unique_id_uses_count=False):
+        aliased_event_name = self._alias_event_name(event_name)
+        return self._emitter.register_last(
+            aliased_event_name, handler, unique_id, unique_id_uses_count
+        )
+
+    def unregister(self, event_name, handler=None, unique_id=None,
+                   unique_id_uses_count=False):
+        aliased_event_name = self._alias_event_name(event_name)
+        return self._emitter.unregister(
+            aliased_event_name, handler, unique_id, unique_id_uses_count
+        )
+
+    def _alias_event_name(self, event_name):
+        if event_name in self._alias_name_cache:
+            return self._alias_name_cache[event_name]
+
+        for old_part, new_part in self._event_aliases.items():
+
+            # We can't simply do a string replace for everything, otherwise we
+            # might end up translating substrings that we never intended to
+            # translate. When there aren't any dots in the old event name
+            # part, then we can quickly replace the item in the list if it's
+            # there.
+            event_parts = event_name.split('.')
+            if '.' not in old_part:
+                try:
+                    # Theoretically a given event name could have the same part
+                    # repeated, but in practice this doesn't happen
+                    event_parts[event_parts.index(old_part)] = new_part
+                except ValueError:
+                    continue
+
+            # If there's dots in the name, it gets more complicated. Now we
+            # have to replace multiple sections of the original event.
+            elif old_part in event_name:
+                old_parts = old_part.split('.')
+                self._replace_subsection(event_parts, old_parts, new_part)
+            else:
+                continue
+
+            new_name = '.'.join(event_parts)
+            logger.debug("Changing event name from %s to %s" % (
+                event_name, new_name
+            ))
+            self._alias_name_cache[event_name] = new_name
+            return new_name
+
+        self._alias_name_cache[event_name] = event_name
+        return event_name
+
+    def _replace_subsection(self, sections, old_parts, new_part):
+        for i in range(len(sections)):
+            if (
+                sections[i] == old_parts[0] and
+                sections[i:i + len(old_parts)] == old_parts
+            ):
+                sections[i:i + len(old_parts)] = [new_part]
+                return
+
+    def __copy__(self):
+        return self.__class__(
+            copy.copy(self._emitter),
+            copy.copy(self._event_aliases)
+        )
+
+
 class _PrefixTrie(object):
     """Specialized prefix trie that handles wildcards.
 
@@ -422,8 +523,9 @@ class _PrefixTrie(object):
                 # will result in final_list = [3, 2, 1], which is
                 # why we reverse the lists.
                 node_list = current_node['values']
-                complete_order = (node_list.first + node_list.middle +
-                                  node_list.last)
+                complete_order = (
+                    node_list.first + node_list.middle + node_list.last
+                )
                 collected.extendleft(reversed(complete_order))
             if not index == key_parts_len:
                 children = current_node['children']
